#ifndef AOCPP_GRID_HPP_
#define AOCPP_GRID_HPP_

#include <iostream>
#include <stdexcept>
#include <type_traits>
#include <utility>
#include <vector>

#include <aocpp/Coord.hpp>

namespace aocpp {

struct Grid;

struct GridElement {
  const Coord pos;
  char elt;
};

class GridIterator {
  Grid const& grid_;
  std::int64_t x_, y_;

  public:
  GridIterator(Grid const& grid, std::int64_t x, std::int64_t y)
  : grid_{grid}, x_{x}, y_{y} {}

  auto operator++() -> GridIterator&;

  auto operator*() -> GridElement;

  auto operator==(GridIterator const& rhs) const -> bool
  {
    return x_ == rhs.x_ && y_ == rhs.y_;
  }
};

struct Grid {
  std::vector<std::string> rows;
  std::size_t columns;
  
  Grid() : rows{}, columns{} {}
  explicit Grid(std::vector<std::string> rows) : rows(std::move(rows))
  {
    columns = rows.empty() ? 0 : rows[0].size();
    for (std::size_t i = 1; i < rows.size(); i++) {
      if (rows[i].size() != columns) {
        throw std::runtime_error{"grid not rectangular"};
      }
    }
  }

  auto Rows() const -> std::size_t { return rows.size(); }
  auto Cols() const -> std::size_t { return columns; }

  auto contains(Coord c) const -> bool {
    return 
         0 <= c.x
      && 0 <= c.y
      && c.y < rows.size()
      && c.x < rows[c.y].size();
  }

  auto operator[](Coord c)       -> char& { return rows[c.y][c.x]; }
  auto operator[](Coord c) const -> char  { return rows[c.y][c.x]; }

  /// @brief Parse a rectangular grid from a stream
  /// @param in input stream read until eof
  /// @return parsed grid
  /// @throw std::runtime\_error when grid is not rectangular
  static auto Parse(std::istream & in) -> Grid {
    Grid result;
    std::string line;
    if (std::getline(in, line)) {
      result.columns = line.size();
      result.rows.emplace_back(std::move(line));
      while (std::getline(in, line)) {
        if (line.size() != result.columns) {
          throw std::runtime_error{"bad grid"};
        }
        result.rows.emplace_back(std::move(line));
      }
    }
    return result;
  }

  auto begin() const -> GridIterator {
    return GridIterator{*this, 0, 0};
  }

  auto end() const -> GridIterator {
    return GridIterator{*this, 0, static_cast<std::int64_t>(rows.size())};
  }
};

}

#endif
